#!/usr/bin/perl

# "caches.pl" in the Forum News Gateway
# copyright 1994,1995 by Swarthmore College

# Manage the cache of recent users.
# Other caches may be added later.

# -- -- -- --

# FlushCaches()
# Write all contents of caches to disk.
# This is done before quitting. The current user should be
# cleared before calling here; this is done by &CleanUpGlobalVars.

sub FlushCaches {
	print STDERR "Flushing caches.\n";
	if ($CurrentUser ne "") {
		print STDERR "BUG: flushing caches, not all needed info is being saved.\n";
	}
	&FlushUserCache();
}

# -- -- -- --

$NCachedUsers = 0;		# the number of filled cache slots
%CachedUsers = ();		# map user names to cache slots

$UserCacheMax = 10;		# max number of users cached - at least 2

# Recent users are cached in variables called
#	%UserNewsrc1				- %Newsrc
#	%UserKillFile1				- %KillFile
#	%UserSubscribed1			- %SubscribedGroups
#	@UserSortedSubscribed1		- @SortedSubscribedGroups
#	$UserTime1					- time that this entry was made in the cache
#	$UserUnsaved1				- whether this user's information needs saving to disk
# and so on for larger numbers, up to $NCachedUsers.

# $NCachedUsers is the number of recent users in the user cache.
# When the gateway starts up it is zero. As more users are cached,
# it rises to $UserCacheMax, where it stays. Each new user cached
# when $NCachedUsers has reached $UserCacheMax replaces an earlier
# user.

# $UserCacheMax should be large enough to accomodate the peak number
# of simultaneous users. Otherwise the cache may thrash, causing extra
# disk activity and recalculation of subscription info--the recalculation
# is particularly slow. It should not be too large, though, because in this
# implementation some operations are linear in the size of the cache.

# If the user cached in a slot is the current user, then the slot's contents
# are invalid; the current data is in the globals %Newsrc and %KillFile.
# When the current user changes, the slot (it's always the same slot) will be
# filled in with the valid data again. We have to be careful not to operate on
# invalid cache slots. That's why &FlushCaches checks that $CurrentUser eq "",
# and why &LazyUserCacheWrite calls &SetCurrentUser("").


# PrintUserCache()
# A debugging routine to print out the state of the user cache.

sub PrintUserCache {
	local(@Users) = ();
	local($User,$Slot);
	
#	print STDERR "-- User Cache\n";
	
	while (($User,$Slot) = each %CachedUsers) {
		unshift(@Users,$User);
	}
	@Users = sort (ByUserCacheSlot @Users);
	
	foreach $User (@Users) {
		$Slot = $CachedUsers{$User};
#		print STDERR "  $Slot. $User saved=",
#			!eval("\$UserUnsaved$Slot"),", time=", eval("\$UserTime$Slot"),"\n";
	}
}

sub ByUserCacheSlot { $CachedUsers{$a} <=> $CachedUsers{$b} }


# AddUserToCache()
# Cache the current user.

sub AddUserToCache {
	local($Slot);				# cache slot number for this user
	
	# Don't cache the empty user.
	return if ($CurrentUser eq "");
	
#	&PrintUserCache();
	
	$Slot = $CachedUsers{$CurrentUser};
	if ($Slot eq "") {
		# This user is not already in the cache.
		if ($NCachedUsers < $UserCacheMax) {
			# If the cache isn't full yet, use the next available slot.
			$NCachedUsers += 1;
			$Slot = $NCachedUsers;
#			print STDERR "New slot $Slot is available.\n";
		} else {
			# The cache is full.
			$Slot = &FreeUserCacheSlot();
#			print STDERR "Reusing old slot $Slot.\n";
		}
		$CachedUsers{$CurrentUser} = $Slot;
	}
	
#	print STDERR "Caching $CurrentUser into slot $Slot.\n";
	
	# A user's Newsrc is timestamped each time the user is made the current
	# user, so \$UserUnsaved$Slot is always true for a newly-cached user.
	eval(
		"\%UserNewsrc$Slot           = \%Newsrc;".
		"\%UserKillFile$Slot         = \%KillFile;".
		"\%UserSubscribed$Slot       = \%SubscribedGroups;".
		"\@UserSortedSubscribed$Slot = \@SortedSubscribedGroups;".
		"\$UserTime$Slot             = time;".
		"\$UserUnsaved$Slot          = 1;"
	);
	
#	&PrintUserCache();
}


# FetchUserFromCache($User)
# If $User is in the cache, make $User the current user and return 1.
# Otherwise return 0.

sub FetchUserFromCache {
	local($User) = @_;
	local($Slot) = 	$CachedUsers{$User};
	
	return 0 if ($Slot eq "");
	
	# A user's Newsrc is timestamped each time the user is made the current
	# user, so there's no point in maintaining a current copy of $UserUnsaved.
	# It would always be true.
	# To save a few bytes, zap the cached values after copying them.
	$CurrentUser = $User;
	eval(
		"\%Newsrc = \%UserNewsrc$Slot;  \%UserNewsrc$Slot = ();".
		"\%KillFile = \%UserKillFile$Slot;  \%UserKillFile$Slot$Slot = ();".
		"\%SubscribedGroups = \%UserSubscribed$Slot;  \%UserSubscribed$Slot = ();".
		"\@SortedSubscribedGroups = \@UserSortedSubscribed$Slot;  \@UserSortedSubscribed$Slot = ();"
	);
	
	return 1;
}


# FreeUserCacheSlot()
# Create a usable cache slot by freeing up an occupied one, and return its index.
# The caller is expected to immediately write into the slot, so what gets
# left behind there is unimportant.
# This is called only when all cache slots are occupied.

sub FreeUserCacheSlot {
	# Find an already-saved slot and return it; if there is none,
	# save and return the oldest unsaved slot.
	local($User,$Slot,$Time);
	local($BestUser);
	local($BestTime) = time+1;			# bigger than any possible time
	foreach $User (keys %CachedUsers) {
		$Slot = $CachedUsers{$User};
#		print STDERR "User $User in slot $Slot is saved: ",!eval("\$UserUnsaved$Slot"),"\n";
		if (!eval("\$UserUnsaved$Slot")) {
			delete $CachedUsers{$User};
			return $Slot;
		}
		$Time = eval("\$UserTime$Slot");
		($BestUser,$BestTime) = ($User,$Time) if ($Time < $BestTime);
	}
	
	# Save the best slot and return it.
	$Slot = $CachedUsers{$BestUser};
	delete $CachedUsers{$BestUser};
#	print STDERR "Emptying a user cache slot by saving $BestUser in slot $Slot.\n";
	eval(
		"&MyDBMClose(\*UserNewsrc$Slot,&NewsrcFilename($BestUser));".
		"&MyDBMClose(\*UserKillFile$Slot,&KillFilename($BestUser));"
	);
	return $Slot;
}


# LazyUserCacheWrite()
# Find the oldest unsaved cache entry, and save it.
# Return 1 if an entry was saved, or 0 if there was no work left to do.

sub LazyUserCacheWrite {
	local($User,$Slot,$Time);
	local($BestUser) = "";
	local($BestTime) = time+1;			# bigger than any possible time
	
	# Clear the current user to put that user's data into the cache, for simplicity.
	&SetCurrentUser("");
	
	while (($User,$Slot) = each %CachedUsers) {
#		print STDERR "User $User in slot $Slot is saved: ",!eval("\$UserUnsaved$Slot"),"\n";
		if (eval("\$UserUnsaved$Slot")) {
			$Time = eval("\$UserTime$Slot");
			($BestUser,$BestTime) = ($User,$Time) if ($Time < $BestTime);
		}
	}
	
	return 0 if ($BestUser eq "");
	
	$Slot = $CachedUsers{$BestUser};
#	print STDERR "Doing lazy flush of user $BestUser in slot $Slot.\n";
	eval(
		"&MyDBMFlush(\*UserNewsrc$Slot,&NewsrcFilename($BestUser));".
		"&MyDBMFlush(\*UserKillFile$Slot,&KillFilename($BestUser));".
		"\$UserUnsaved$Slot = 0;"
	);
	return 1;
}


# FlushUserCache()
# Make sure that everything in the user cache is written to disk.
# This is done before quitting.

sub FlushUserCache {
	local($User);
	local($Slot);
	
	while (($User,$Slot) = each %CachedUsers) {
		eval(
			"if (\$UserUnsaved$Slot) {".
				"&MyDBMClose(\*UserNewsrc$Slot,&NewsrcFilename($User));".
				"&MyDBMClose(\*UserKillFile$Slot,&KillFilename($User));".
			"}"
		);
	}
}

1;
