#!/usr/bin/perl

# "newsrc.pl" in the Forum News Gateway
# copyright 1994,1995 by Swarthmore College

# This file contains routines to manipulate newsrc data.
# Each user has their own newsrc. See "globals.pl" for SetCurrentUser, which
# closes the old newsrc and opens a new one.

# The newsrc contains:

# $Newsrc{"prefs","password"} = "[user password, encrypted if encryption is available]"
# $Newsrc{"prefs","realname"} = "[user's real name]"
# $Newsrc{"prefs","email"} = "[user's e-mail address]"
# $Newsrc{"prefs","organization"} = "[user's organization]"
# $Newsrc{"prefs","markread"} = 1 or 0, whether to mark articles read as the user reads them
# $Newsrc{"prefs","hideempty"} = 1 or 0, whether to hide subscribed groups with no unread articles from home news page
# $Newsrc{"time","created"} = [when account was created, as returned by time()]
# $Newsrc{"time","used"} = [when account was last used, as returned by time()]
# $Newsrc{"[newsrc key]"} = "[newsrc entry]";
# $Newsrc{"[newsrc key]","lastkill"} = "[number of last article scanned for killing]"

# A [newsrc key] has the form '$NNTPServer." ".$CurrentGroup'.
# $NNTPServer is the name of the news server on which the group exists.  The server name is
# in a canonical form, which is the full hostname.  If the port number is not 119 (the standard
# NNTP port), the hostname is followed by a colon and the port number.
# An example key is "larch.cc.swarthmore.edu rec.games.abstract". $Newsrc has a [newsrc key] for each group
# that the user has read.

# A [newsrc entry] may be empty. If not, it is "! " or ": " followed by a list of read articles.
# ": " means the user is subscribed to the group and "! " means the user is not subscribed.
# The list of read articles is a sorted list, in ascending order, of the article numbers of read
# articles. For example, "2,5,9". When the article number are consecutive, they're listed as
# ranges instead of as individual articles. For example, "2,5-10,13,19-20".

# -- --

# HasArticleBeenRead($ArticleNumber)
# Returns true if the given article in the current group is marked as read in the newsrc.

sub HasArticleBeenRead {
	local($ArticleNumber) = @_;

	local($NewsrcKey) = $CurrentNNTPServer." ".$CurrentGroup;
	local($NewsrcEntry) = $Newsrc{$NewsrcKey};
	local($ArticlesReadNewsrcPart) = "";
	
	if ($NewsrcEntry ne "") {
		if ($NewsrcEntry !~ /^(!|:)\s*([\d\-\,]*)/) {
			print STDERR "In HasArticleBeenRead: User $CurrentUser, Newsrc entry \"$NewsrcKey\" is bad: \"$Newsrc{$NewsrcKey}\"\n";
			return 0;
		}
		$ArticlesReadNewsrcPart = $2;
		
		local(@Ranges) = split(/\,/,$ArticlesReadNewsrcPart);
		local($Range);
		local($RangeLowEnd,$RangeHighEnd);
		foreach $Range (@Ranges) {
			if ($Range !~ /^(\d+)(\-(\d+))?/) {
				print STDERR "In HasArticleBeenRead: User $CurrentUser, Newsrc entry \"$NewsrcKey\" has bad range: \"$Range\"\n";
				return 0;
			}
			$RangeLowEnd = $1;
			if ($3 ne "") {
				$RangeHighEnd = $3;
			} else {
				$RangeHighEnd = $RangeLowEnd;
			}
			
			# If we're past the article number, it is unread.
			return 0 if ($RangeLowEnd > $ArticleNumber);
			
			# If we haven't gotten to the article number, go to the next range.
			next if ($RangeHighEnd < $ArticleNumber);
			
			# It must be true that ($RangeLowEnd <= $ArticleNumber && $RangeHighEnd >= $ArticleNumber),
			# so it's in this range, and it has been read
			return 1;
		}
	}
	
	# It has not been read.
	return 0;
}

# MarkArticleRead($ArticleNum)
# Update the newsrc entry for the current group to show that the given article has been read.

sub MarkArticleRead {
	local($ArticleNum) = @_;

	if ($ArticleNum < 1) {
		print STDERR "In MarkArticleRead: \$ArticleNum < 1\n";
		return;
	}
	
	local($NewsrcKey) = $CurrentNNTPServer." ".$CurrentGroup;
	local($NewsrcEntry) = $Newsrc{$NewsrcKey};
#	print STDERR "Entering MarkArticleRead: About to mark article $ArticleNum read.\n";
#	print STDERR "    Newsrc entry for \"$NewsrcKey\" is \"$Newsrc{$NewsrcKey}\"\n";
	
	if ($NewsrcEntry ne "") {
		local($SubscribedCharNewsrcPart) = "";
		local($ArticlesReadNewsrcPart) = "";
		
		if ($NewsrcEntry !~ /^(!|:)\s*([\d\-\,]*)/) {
			print STDERR "In MarkArticleRead: User $CurrentUser, Newsrc entry \"$NewsrcKey\" is bad: \"$Newsrc{$NewsrcKey}\"\n";
			return;
		}
		$SubscribedCharNewsrcPart = $1;
		$ArticlesReadNewsrcPart = $2;
		
		local(@Ranges) = split(/\,/,$ArticlesReadNewsrcPart);
		local($CurrRangeLow,$CurrRangeHigh,$PrevRangeLow,$PrevRangeHigh,$I);
		local($Marked) = 0;
		$PrevRangeLow = undef;
		$PrevRangeHigh = undef;
		for ($I = 0; $I <= $#Ranges; $I++) {
			if ($Ranges[$I] !~ /^(\d+)(\-(\d+))?/) {
				print STDERR "In MarkArticleRead: User $CurrentUser, Newsrc entry \"$NewsrcKey\" has bad range: \"$Ranges[$I]\"; erasing this range\n";
				$Ranges[$I] = "";
				next;
			}
			$CurrRangeLow = $1;
			if ($3 ne "") {
				$CurrRangeHigh = $3;
			} else {
				$CurrRangeHigh = $CurrRangeLow;
			}
#			print STDERR "In MarkArticleRead: CurrRange is $CurrRangeLow-$CurrRangeHigh\n";
			
			if ($CurrRangeLow <= $ArticleNum && $CurrRangeHigh >= $ArticleNum) {
				$Marked = 1; 	# already read
				last;
			}
			if ($CurrRangeLow > $ArticleNum) {
				# we've found the first range that is higher than the given article.  Figure out
				# if we have to coalesce any ranges.
				local($StickOnToPrevRange) = defined($PrevRangeHigh) && ($PrevRangeHigh == $ArticleNum-1);
				local($StickOnToCurrRange) = ($CurrRangeLow == $ArticleNum+1);
				
#				print STDERR "StickOnToPrevRange is $StickOnToPrevRange\n";
#				print STDERR "StickOnToCurrRange is $StickOnToCurrRange\n";

				if ($StickOnToPrevRange && $StickOnToCurrRange) {
					# Delete previous range and make current range include it
					$Ranges[$I-1] = "";
					$Ranges[$I] = "$PrevRangeLow-$CurrRangeHigh";
				} elsif ($StickOnToPrevRange) {
					$Ranges[$I-1] = "$PrevRangeLow-$ArticleNum";
				} elsif ($StickOnToCurrRange) {
					$Ranges[$I] = "$ArticleNum-$CurrRangeHigh";
				} else {
					# We have to create a new range.  All the ranges will be concatted together with commas
					# at the end of this routine, so I just stick another range on to the current range.
					$Ranges[$I] = $ArticleNum.','.$Ranges[$I];
				}
				$Marked = 1;
				last;
			}
			$PrevRangeLow = $CurrRangeLow;
			$PrevRangeHigh = $CurrRangeHigh;
		}
		
		# Now add it if we didn't somewhere in the loop.
		if (!$Marked) {
			local($StickOnToLastRange) = defined($PrevRangeHigh) && ($PrevRangeHigh == $ArticleNum-1);
			
			if ($StickOnToLastRange) {
				$Ranges[$#Ranges] = "$PrevRangeLow-$ArticleNum";
			} else {
				$Ranges[$#Ranges+1] = "$ArticleNum";
			}
		}
		
		# Join up the ranges and delete empty ones.
		$ArticlesReadNewsrcPart = join(',',@Ranges);
		$ArticlesReadNewsrcPart =~ s/^,+//;
		$ArticlesReadNewsrcPart =~ s/,+$//;
		$ArticlesReadNewsrcPart =~ s/,,+/,/g;
		
		$Newsrc{$NewsrcKey} = $SubscribedCharNewsrcPart." ".$ArticlesReadNewsrcPart;
	} else {
		$Newsrc{$NewsrcKey} = "! $ArticleNum";
	}
	
#	print STDERR "Exiting MarkArticleRead: Newsrc entry for \"$NewsrcKey\" is \"$Newsrc{$NewsrcKey}\"\n";
}

# MarkArticleUnread($ArticleNum)
# Update the newsrc entry for the current group to show that the given article has not been read.

sub MarkArticleUnread {
	local($ArticleNum) = @_;
	
	if ($ArticleNum < 1) {
		print STDERR "In MarkArticleUnread: \$ArticleNum < 1\n";
		return;
	}
	
	local($NewsrcKey) = $CurrentNNTPServer." ".$CurrentGroup;
	local($NewsrcEntry) = $Newsrc{$NewsrcKey};
#	print STDERR "Entering MarkArticleUnread: About to mark article $ArticleNum unread.\n";
#	print STDERR "    Newsrc entry for \"$NewsrcKey\" is \"$Newsrc{$NewsrcKey}\"\n";

	if ($NewsrcEntry ne "") {
		local($SubscribedCharNewsrcPart) = "";
		local($ArticlesReadNewsrcPart) = "";
		
		if ($NewsrcEntry !~ /^(!|:)\s*([\d\-\,]*)/) {
			print STDERR "In MarkArticleUnread: User $CurrentUser, Newsrc entry \"$NewsrcKey\" is bad: \"$Newsrc{$NewsrcKey}\"\n";
			return;
		}
		$SubscribedCharNewsrcPart = $1;
		$ArticlesReadNewsrcPart = $2;
		
		local(@Ranges) = split(/\,/,$ArticlesReadNewsrcPart);
		local($CurrRangeLow,$CurrRangeHigh,$I);
		for ($I = 0; $I <= $#Ranges; $I++) {
			if ($Ranges[$I] !~ /^(\d+)(\-(\d+))?/) {
				print STDERR "In MarkArticleUnread: User $CurrentUser, Newsrc entry \"$NewsrcKey\" has bad range: \"$Ranges[$I]\"; erasing this range\n";
				$Ranges[$I] = "";
				next;
			}
			$CurrRangeLow = $1;
			if ($3 ne "") {
				$CurrRangeHigh = $3;
			} else {
				$CurrRangeHigh = $CurrRangeLow;
			}
#			print STDERR "In MarkArticleUnread: CurrRange is $CurrRangeLow-$CurrRangeHigh\n";
			
			if ($CurrRangeLow <= $ArticleNum && $CurrRangeHigh >= $ArticleNum) {
				if ($CurrRangeLow == $ArticleNum && $CurrRangeHigh == $ArticleNum) {
					$Ranges[$I] = "";
				} elsif ($CurrRangeLow == $ArticleNum) {
					$CurrRangeLow++;
					if ($CurrRangeLow == $CurrRangeHigh) {
						$Ranges[$I] = "$CurrRangeLow";
					} else {
						$Ranges[$I] = "$CurrRangeLow-$CurrRangeHigh";
					}
				} elsif ($CurrRangeHigh == $ArticleNum) {
					$CurrRangeHigh--;
					if ($CurrRangeLow == $CurrRangeHigh) {
						$Ranges[$I] = "$CurrRangeLow";
					} else {
						$Ranges[$I] = "$CurrRangeLow-$CurrRangeHigh";
					}
				} else {
					# Break the current range in two, appending the new one with a ",".
					if ($CurrRangeLow == $ArticleNum-1) {
						$Ranges[$I] = "$CurrRangeLow";
					} else {
						local($Temp) = $ArticleNum-1;
						$Ranges[$I] = "$CurrRangeLow-$Temp";
					}
					if ($CurrRangeHigh == $ArticleNum+1) {
						$Ranges[$I] .= ",$CurrRangeHigh";
					} else {
						local($Temp) = $ArticleNum+1;
						$Ranges[$I] .= ",$Temp-$CurrRangeHigh";
					}
				}
				last;
			}
		}
		
		# Join up the ranges and delete empty ones.
		$ArticlesReadNewsrcPart = join(',',@Ranges);
		$ArticlesReadNewsrcPart =~ s/^,+//;
		$ArticlesReadNewsrcPart =~ s/,+$//;
		$ArticlesReadNewsrcPart =~ s/,,+/,/g;

		$Newsrc{$NewsrcKey} = $SubscribedCharNewsrcPart." ".$ArticlesReadNewsrcPart;
	}
	
#	print STDERR "Exiting MarkArticleUnread: Newsrc entry for \"$NewsrcKey\" is \"$Newsrc{$NewsrcKey}\"\n";
}

# MarkReadUpTo($ArticleNum)
# Update the newsrc entry for the current group to show that all articles from 1-$ArticleNum have been read.
# This is used to mark read all the articles which no longer exist on the news server.

sub MarkReadUpTo {
	local($ArticleNum) = @_;

	return if ($ArticleNum < 1);
	
	local($NewsrcKey) = $CurrentNNTPServer." ".$CurrentGroup;
	local($NewsrcEntry) = $Newsrc{$NewsrcKey};
#	print STDERR "Entering MarkReadUpTo: About to mark articles 1-$ArticleNum read.\n";
#	print STDERR "    Newsrc entry for \"$NewsrcKey\" is \"$Newsrc{$NewsrcKey}\"\n";
	
	if ($NewsrcEntry ne "") {
		local($SubscribedCharNewsrcPart) = "";
		local($ArticlesReadNewsrcPart) = "";
		
		if ($NewsrcEntry !~ /^(!|:)\s*([\d\-\,]*)/) {
			print STDERR "In MarkReadUpTo: User $CurrentUser, Newsrc entry \"$NewsrcKey\" is bad: \"$Newsrc{$NewsrcKey}\"\n";
			return;
		}
		$SubscribedCharNewsrcPart = $1;
		$ArticlesReadNewsrcPart = $2;
		
		local(@Ranges) = split(/\,/,$ArticlesReadNewsrcPart);
		local($CurrRangeLow,$CurrRangeHigh,$I);
		local($Marked) = 0;
		for ($I = 0; $I <= $#Ranges; $I++) {
			if ($Ranges[$I] !~ /^(\d+)(\-(\d+))?/) {
				print STDERR "In MarkReadUpTo: User $CurrentUser, Newsrc entry \"$NewsrcKey\" has bad range: \"$Ranges[$I]\"; erasing this range\n";
				$Ranges[$I] = "";
				next;
			}
			$CurrRangeLow = $1;
			if ($3 ne "") {
				$CurrRangeHigh = $3;
			} else {
				$CurrRangeHigh = $CurrRangeLow;
			}
#			print STDERR "In MarkReadUpTo: CurrRange is $CurrRangeLow-$CurrRangeHigh\n";
			
			if ($CurrRangeHigh < $ArticleNum) {
				$Ranges[$I] = "";
			} else {
				if ($CurrRangeLow <= $ArticleNum+1) {
					$ArticleNum = $CurrRangeHigh;
					$Ranges[$I] = "";
				} else {
					if ($ArticleNum == 1) {
						$Ranges[$I] = "1,".$Ranges[$I];
					} else {
						$Ranges[$I] = "1-$ArticleNum,".$Ranges[$I];
					}
					$Marked = 1;
					last;
				}
			}
		}
		
		# Now add it if we didn't somewhere in the loop
		# Add it at the end since if we didn't add it in the loop then
		# we deleted all the ranges that used to be there (or made them "", anyway)
		if (!$Marked) {
			if ($ArticleNum == 1) {
				$Ranges[$#Ranges+1] = "1";
			} else {
				$Ranges[$#Ranges+1] = "1-$ArticleNum";
			}
		}
		
		# Join up the ranges and delete empty ones.
		$ArticlesReadNewsrcPart = join(',',@Ranges);
		$ArticlesReadNewsrcPart =~ s/^,+//;
		$ArticlesReadNewsrcPart =~ s/,+$//;
		$ArticlesReadNewsrcPart =~ s/,,+/,/g;
		
		$Newsrc{$NewsrcKey} = $SubscribedCharNewsrcPart." ".$ArticlesReadNewsrcPart;
	} else {
		$Newsrc{$NewsrcKey} = "! 1-$ArticleNum";
	}
	
#	print STDERR "Exiting MarkReadUpTo: Newsrc entry for \"$NewsrcKey\" is \"$Newsrc{$NewsrcKey}\"\n";
}

# MarkThreadRead($Subject)
# Update the newsrc entry for the current group to show that all articles in the given
# thread (specified by its canonical subject) have been read.

sub MarkThreadRead {
	local($Subject) = @_;
	local(@Roots) = split(/\t/,$ThreadDB{"threadroots",$Subject});
	local($Root);
	
	foreach $Root (@Roots) {
		&MarkSubthreadRead($Root);
	}
}

# MarkSubthreadRead($Root)
# Update the newsrc entry for the current group to show that all articles in the given
# subthread (specified by its root article) have been read.

sub MarkSubthreadRead {
	local($Root) = @_;
	local($ArticleNum) = $ThreadDB{$Root,"artnum"};
	
	&MarkArticleRead($ArticleNum) if ($ArticleNum ne "dummy");
	
	local($Children) = $ThreadDB{$Root,"children"};
	local($Child);
	foreach $Child (split(/ /,$Children)) {
		&MarkSubthreadRead($Child);
	}
}

# NumUnreadArticlesInGroup()
# Returns the number of articles in the current group which
# have not yet been read.

sub NumUnreadArticlesInGroup {
	local($LowArticleNumber,$HighArticleNumber) = ($CurrentGroupLowWaterMark,$CurrentGroupHighWaterMark);

	local($NumArticlesUnread);
	if ($HighArticleNumber >= $LowArticleNumber && $LowArticleNumber > 0) {
		$NumArticlesUnread = $HighArticleNumber - $LowArticleNumber + 1;
	} else {
		$NumArticlesUnread = 0;
	}
	
	local($NewsrcKey) = $CurrentNNTPServer." ".$CurrentGroup;
	local($NewsrcEntry) = $Newsrc{$NewsrcKey};
	local($ArticlesReadNewsrcPart);
	if ($NewsrcEntry ne "") {
		if ($NewsrcEntry !~ /^(!|:)\s*([\d\-\,]*)/) {
			print STDERR "In NumUnreadArticlesInGroup: User $CurrentUser, Newsrc entry \"$NewsrcKey\" is bad: \"$Newsrc{$NewsrcKey}\"\n";
			return 0;
		}
		$ArticlesReadNewsrcPart = $2;
		
		local(@Ranges) = split(/\,/,$ArticlesReadNewsrcPart);
		local($Range);
		local($RangeLowEnd,$RangeHighEnd);
		foreach $Range (@Ranges) {
			if ($Range !~ /^(\d+)(\-(\d+))?/) {
				print STDERR "In NumUnreadArticlesInGroup: User $CurrentUser, Newsrc entry \"$NewsrcKey\" has bad range: \"$Range\"\n";
				return 0;
			}
			$RangeLowEnd = $1;
			if ($3 ne "") {
				$RangeHighEnd = $3;
			} else {
				$RangeHighEnd = $RangeLowEnd;
			}
			
			# If we're past the range of articles that exist, stop and return.
			last if ($RangeLowEnd > $HighArticleNumber);
				
			# If we haven't gotten to the range of articles that exist, go to the next range.
			next if ($RangeHighEnd < $LowArticleNumber);
			
			# It must be true that ($RangeLowEnd <= $HighArticleNumber && $RangeHighEnd >= $LowArticleNumber),
			# so this range overlaps with the range of articles that exist. Subtract the number of articles
			# in the overlap.
			local($OverlapLow) = ($RangeLowEnd >= $LowArticleNumber) ? $RangeLowEnd : $LowArticleNumber;
			local($OverlapHigh) = ($RangeHighEnd <= $HighArticleNumber) ? $RangeHighEnd : $HighArticleNumber;
			$NumArticlesUnread -= ($OverlapHigh-$OverlapLow+1);
		}
	}
	
	return $NumArticlesUnread;
}

# UpdateSubscriptionInfo()
# Updates the global %SubscribedGroups and @SortedSubscribedGroups, if necessary.
# %SubscribedGroups caches the number of unread articles in each group, for the current user.
# @SortedSubscribedGroups is a list of the groups in order.
# The variables are cleared whenever something might invalidate them, and
# set only by this routine. If you want 'em, call here first.

sub UpdateSubscriptionInfo {
	local($NewsrcKey,$NewsrcEntry);
	
	# If they're already set, don't set them again.
	return if (@SortedSubscribedGroups > 0);
	
	%SubscribedGroups = ();
	
	local (@Entries) = ();
	
	# Collect the list of subscribed newsrc entries into @Entries.
	while (($NewsrcKey,$NewsrcEntry) = each %Newsrc) {
		unshift(@Entries,$NewsrcKey) if ($NewsrcEntry =~ /^:/);
	}
	
	# Check each group for existence.
	# For efficiency, we try to change the current news server as rarely as possible.
	local ($Server);
	local ($AnyServer);				# whether the news server has been set
	local ($Entry);					# a newsrc entry
	local ($Server,$Group);
	local ($OK);					# whether the news server seems to be OK
	foreach $Server (@CanonicalNNTPServers) {
		$AnyServer = 0;
		foreach $Entry (@Entries) {
			local($GroupServer,$GroupName) = split(/ /,$Entry);
			if ($GroupServer eq $Server) {
				if (!$AnyServer) {
					$AnyServer = 1;
					eval <<'EndOfEval';
						&SetCurrentNNTPServer($GroupServer,0);
EndOfEval
					if ($ExceptionCode) {
						$ExceptionCode = 0;
						$ExceptionText = "";
						$OK = 0;					# the news server seems to be hosed
					} else {
						$OK = 1;					# the news server seems to be OK
					}
				}
				if ($OK) {
					eval <<'EndOfEval';
						&SetCurrentGroup($GroupName,0);
EndOfEval
					if ($ExceptionCode) {
						$ExceptionCode = 0;
						$ExceptionText = "";
						# Don't set $SubscribedGroups{$Entry}, so the group will not appear.
					} else {
						$SubscribedGroups{$Entry} = &NumUnreadArticlesInGroup();
					}
				}
			}
		}
	}
	
	@SortedSubscribedGroups = sort NewsrcKeyCompare keys(%SubscribedGroups);
}

# ClearSubscriptionCache()
# Clear the cache of subscribed newsgroup info.

sub ClearSubscriptionCache {
#	print STDERR "Clearing subscription cache.\n";
	undef %SubscribedGroups;
	undef @SortedSubscribedGroups;
	@SortedSubscribedGroups = ();
}

# SubscribeToGroup()
# Update the newsrc to show that the current group is subscribed to.

sub SubscribeToGroup {
	local($NewsrcKey) = $CurrentNNTPServer." ".$CurrentGroup;

	local($ArticlesRead) = "";
	if ($Newsrc{$NewsrcKey} ne "") {
		if ($Newsrc{$NewsrcKey} !~ /^(!|:)\s*([\d,\-]*)?/) {
			print STDERR "In SubscribeToGroup: User $CurrentUser, Newsrc entry \"$NewsrcKey\" is bad: \"$Newsrc{$NewsrcKey}\".\n";
			return;
		}
		$ArticlesRead = $2;
	}
	
	$Newsrc{$NewsrcKey} = ': '.$ArticlesRead;
	
	&ClearSubscriptionCache();
}

# UnsubscribeFromGroup()
# Update the newsrc to show that the current group is not subscribed to.

sub UnsubscribeFromGroup {
	local($NewsrcKey) = $CurrentNNTPServer." ".$CurrentGroup;

	local($ArticlesRead) = "";
	if ($Newsrc{$NewsrcKey} ne "") {
		if ($Newsrc{$NewsrcKey} !~ /^(!|:)\s*([\d,\-]*)?/) {
			print STDERR "In UnsubscribeFromGroup: User $CurrentUser, Newsrc entry \"$NewsrcKey\" is bad: \"$Newsrc{$NewsrcKey}\".\n";
			return;
		}
		$ArticlesRead = $2;
	}
	
	$Newsrc{$NewsrcKey} = '! '.$ArticlesRead;
	
	&ClearSubscriptionCache();
}

# MarkGroupRead()
# Update the newsrc to show that all articles in the current group have been read.

sub MarkGroupRead {
	local($LowArticleNumber,$HighArticleNumber) = ($ThreadDB{"DBRange","low"},$ThreadDB{"DBRange","high"});
	
	local($NewsrcKey) = $CurrentNNTPServer." ".$CurrentGroup;
	
	local($SubscribedChar) = "!"; 		# unsubscribed until proven otherwise
	if ($Newsrc{$NewsrcKey} ne "") {
		if ($Newsrc{$NewsrcKey} !~ /^(!|:)/) {
			print STDERR "In  MarkGroupRead: User $CurrentUser, Newsrc entry \"$NewsrcKey\" is bad: \"$Newsrc{$NewsrcKey}\".\n";
			return;
		}
		$SubscribedChar = $1;
	}
	
	$Newsrc{$NewsrcKey} = $SubscribedChar;
	if ($HighArticleNumber > 0) {
		$Newsrc{$NewsrcKey} .= ' 1-'.$HighArticleNumber;
	}
	
	# If subscription info is cached, and we're subscribed to this group, then
	# set the numunread for this group to 0.
	# This is lots more efficient than zapping the whole cache.
	if (@SortedSubscribedGroups > 0 && $SubscribedGroups{$NewsrcKey} != 0) {
		$SubscribedGroups{$NewsrcKey} = 0;
	}
}

# MarkGroupUnread()
# Update the newsrc to show that all articles in the current group have not been read.

sub MarkGroupUnread {
	local($LowArticleNumber,$HighArticleNumber) = ($ThreadDB{"DBRange","low"},$ThreadDB{"DBRange","high"});
	
	local($NewsrcKey) = $CurrentNNTPServer." ".$CurrentGroup;
	
	local($SubscribedChar) = "!"; 		# unsubscribed until proven otherwise
	if ($Newsrc{$NewsrcKey} ne "") {
		if ($Newsrc{$NewsrcKey} !~ /^(!|:)/) {
			print STDERR "In MarkGroupUnread: User $CurrentUser, Newsrc entry \"$NewsrcKey\" is bad: \"$Newsrc{$NewsrcKey}\".\n";
			return;
		}
		$SubscribedChar = $1;
	}
	
	$Newsrc{$NewsrcKey} = $SubscribedChar;
	if ($LowArticleNumber > 1) {
		$Newsrc{$NewsrcKey} .= ' 1-'.($LowArticleNumber-1);
	}
	
	&ClearSubscriptionCache();
}

1;
